# Copyright OSISM GmbH, 2022-2023
# LICENSE: CC BY-NC 4.0

from osism import utils


def update_network_interface_name(mac_address, network_interface_name):
    """Sets the network interface name from a performed
    introspection for an interface with a given MAC address."""

    interface_a = utils.nb.dcim.interfaces.get(mac_address=mac_address)
    interface_a.custom_fields = {"network_interface_name": network_interface_name}
    interface_a.save()



import datetime
from logging.config import dictConfig
import logging
from uuid import UUID

from fastapi import FastAPI, Header, Request, Response
from pydantic import BaseModel
import pynetbox
from starlette.middleware.cors import CORSMiddleware

from osism.tasks import reconciler
from osism import settings


class WebhookNetboxResponse(BaseModel):
    result: str


class WebhookNetboxData(BaseModel):
    username: str
    data: dict
    snapshots: dict
    event: str
    timestamp: datetime.datetime
    model: str
    request_id: UUID


# https://stackoverflow.com/questions/63510041/adding-python-logging-to-fastapi-endpoints-hosted-on-docker-doesnt-display-api
class LogConfig(BaseModel):
    """Logging configuration to be set for the server"""

    LOGGER_NAME: str = "mycoolapp"
    LOG_FORMAT: str = "%(levelprefix)s | %(asctime)s | %(message)s"
    LOG_LEVEL: str = "DEBUG"

    # Logging config
    version = 1
    disable_existing_loggers = False
    formatters = {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": LOG_FORMAT,
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    }
    handlers = {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
    }
    loggers = {
        "api": {"handlers": ["default"], "level": LOG_LEVEL},
    }


app = FastAPI()

app.add_middleware(CORSMiddleware)

dictConfig(LogConfig().dict())
logger = logging.getLogger("api")

nb = None


@app.on_event("startup")
async def startup_event():
    global nb

    nb = pynetbox.api(settings.NETBOX_URL, token=settings.NETBOX_TOKEN)

    if settings.IGNORE_SSL_ERRORS:
        import requests

        requests.packages.urllib3.disable_warnings()
        session = requests.Session()
        session.verify = False
        nb.http_session = session


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.post("/webhook/netbox", response_model=WebhookNetboxResponse, status_code=200)
async def webhook(
    webhook_input: WebhookNetboxData,
    request: Request,
    response: Response,
    content_length: int = Header(...),
    x_hook_signature: str = Header(None),
):
    global nb

    data = webhook_input.data
    url = data["url"]
    name = data["name"]

    if "devices" in url:
        tags = [x["name"] for x in data["tags"]]

        custom_fields = data["custom_fields"]
        device_type = custom_fields["device_type"]

        # NOTE: device without a defined device_type are nodes
        if not device_type:
            device_type = "node"

    elif "interfaces" in url:
        device_type = "interface"

        device_id = data["device"]["id"]
        device = nb.dcim.devices.get(id=device_id)
        tags = [str(x) for x in device.tags]
        custom_fields = device.custom_fields

    if "Managed by OSISM" in tags:
        if device_type == "server":
            logger.info(
                f"Handling change for managed device {name} of type {device_type}"
            )
            reconciler.run.delay()
        elif device_type == "switch":
            logger.info(
                f"Handling change for managed device {name} of type {device_type}"
            )
            # netbox.generate.delay(name, custom_fields['configuration_template'])
        elif device_type == "interface":
            logger.info(
                f"Handling change for interface {name} on managed device {device.name} of type {custom_fields['device_type']}"
            )
            # netbox.generate.delay(device.name, custom_fields['configuration_template'])

    else:
        logger.info(f"Ignoring change for unmanaged device {name}")

    return {"result": "ok"}



from datetime import datetime

from celery import Celery
from cliff.command import Command
from redis import Redis
from tabulate import tabulate

from osism.tasks import Config


class List(Command):
    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)
        parser.add_argument(
            "--status", default="all", help="Status of the tasks to list"
        )
        return parser

    def take_action(self, parsed_args):
        status = parsed_args.status
        redis = Redis(host="redis", port="6379")

        app = Celery("task")
        app.config_from_object(Config)

        i = app.control.inspect()

        table = []

        task_status = "ACTIVE"
        for worker, tasks in i.active().items():
            for task in tasks:
                time_start = datetime.fromtimestamp(task["time_start"])
                table.append(
                    [
                        worker,
                        task["id"],
                        task["name"],
                        task_status,
                        time_start,
                        task["args"],
                    ]
                )

        task_status = "SCHEDULED"
        for worker, tasks in i.scheduled().items():
            for task in tasks:
                time_start = datetime.fromtimestamp(task["time_start"])
                table.append(
                    [
                        worker,
                        task["id"],
                        task["name"],
                        task_status,
                        time_start,
                        task["args"],
                    ]
                )

        print(
            tabulate(
                table,
                headers=["Worker", "ID", "Name", "Status", "Start time", "Arguments"],
                tablefmt="psql",
            )
        )


class Revoke(Command):
    def get_parser(self, prog_name):
        parser = super(Revoke, self).get_parser(prog_name)
        parser.add_argument("task", nargs=1, type=str, help="Task to revoke")
        return parser

    def take_action(self, parsed_args):
        task = parsed_args.task

        app = Celery("task")
        app.config_from_object(Config)
        app.control.revoke(task, terminate=True)



LOADBALANCER_PLAYBOOKS = [
    "loadbalancer-aodh",
    "loadbalancer-barbican",
    # "loadbalancer-blazar",
    # "loadbalancer-ceph-rgw",
    "loadbalancer-cinder",
    # "loadbalancer-cloudkitty",
    # "loadbalancer-cyborg",
    "loadbalancer-designate",
    "loadbalancer-elasticsearch",
    # "loadbalancer-freezer",
    "loadbalancer-glance",
    "loadbalancer-gnocchi",
    "loadbalancer-grafana",
    "loadbalancer-heat",
    "loadbalancer-horizon",
    # "loadbalancer-influxdb",
    "loadbalancer-ironic",
    "loadbalancer-keystone",
    "loadbalancer-kibana",
    # "loadbalancer-magnum",
    "loadbalancer-manila",
    "loadbalancer-mariadb",
    # "loadbalancer-masakari",
    "loadbalancer-memcached",
    # "loadbalancer-mistral",
    # "loadbalancer-monasca",
    # "loadbalancer-murano",
    "loadbalancer-neutron",
    "loadbalancer-nova",
    "loadbalancer-octavia",
    "loadbalancer-placement",
    "loadbalancer-prometheus",
    "loadbalancer-rabbitmq",
    # "loadbalancer-sahara",
    # "loadbalancer-senlin",
    "loadbalancer-skydive",
    # "loadbalancer-solum",
    # "loadbalancer-swift",
    # "loadbalancer-tacker",
    # "loadbalancer-trove",
    # "loadbalancer-vitrage",
    # "loadbalancer-watcher",
    # "loadbalancer-zun",
]

VALIDATE_PLAYBOOKS = {
    "barbican-config": {
        "runtime": "kolla-ansible",
        "playbook": "barbican",
    },
    "designate-config": {
        "runtime": "kolla-ansible",
        "playbook": "designate",
    },
    "keystone-config": {
        "runtime": "kolla-ansible",
        "playbook": "keystone",
    },
    "glance-config": {
        "runtime": "kolla-ansible",
        "playbook": "glance",
    },
    "heat-config": {
        "runtime": "kolla-ansible",
        "playbook": "heat",
    },
    "octavia-config": {
        "runtime": "kolla-ansible",
        "playbook": "octavia",
    },
    "nova-config": {
        "runtime": "kolla-ansible",
        "playbook": "nova",
    },
    "neutron-config": {
        "runtime": "kolla-ansible",
        "playbook": "neutron",
    },
    "placement-config": {
        "runtime": "kolla-ansible",
        "playbook": "placement",
    },
    "aodh-config": {
        "runtime": "kolla-ansible",
        "playbook": "aodh",
    },
    "ceilometer-config": {
        "runtime": "kolla-ansible",
        "playbook": "ceilometer",
    },
    "cloudkitty-config": {
        "runtime": "kolla-ansible",
        "playbook": "cloudkitty",
    },
    "ironic-config": {
        "runtime": "kolla-ansible",
        "playbook": "ironic",
    },
    "manila-config": {
        "runtime": "kolla-ansible",
        "playbook": "manila",
    },
    "senlin-config": {
        "runtime": "kolla-ansible",
        "playbook": "senlin",
    },
    # NOTE: The command should be "osism validate ceph-config". However,
    # the corresponding playbook is called ceph-validate because ceph-config
    # deploys the Ceph configuration itself. So this is rewritten from
    # ceph-config to ceph-validate.
    "ceph-config": {
        "runtime": "ceph-ansible",
        "playbook": "validate",
    },
    # NOTE: The playbooks for validating the Ceph deployment are currently
    # in osism/ansible-playbooks. Therefore, they are not executed in
    # ceph-ansible but in osism-ansible.
    "ceph-mgrs": {"environment": "ceph", "runtime": "osism-ansible"},
    "ceph-mons": {"environment": "ceph", "runtime": "osism-ansible"},
    "ceph-osds": {"environment": "ceph", "runtime": "osism-ansible"},
    "container-status": {"environment": "generic", "runtime": "osism-ansible"},
    "kernel-version": {"environment": "generic", "runtime": "osism-ansible"},
    "mysql-open-files-limit": {"environment": "generic", "runtime": "osism-ansible"},
    "refstack": {"environment": "openstack", "runtime": "osism-ansible"},
    "system-encoding": {"environment": "generic", "runtime": "osism-ansible"},
    "ulimits": {"environment": "generic", "runtime": "osism-ansible"},
}

MAP_ROLE2ROLE = {
    "ceph-basic": [
        "ceph-infra",
        "ceph-mons",
        "ceph-mgrs",
        "ceph-osds",
        "ceph-crash",
    ],
    "infrastructure-basic": [
        "openstackclient",
        "common",
        "loadbalancer",
        "elasticsearch",
        "kibana",
        "openvswitch",
        "memcached",
        "redis",
        "mariadb",
        "rabbitmq",
        "phpmyadmin",
    ],
    "openstack-basic": [
        "keystone",
        "horizon",
        "placement",
        "glance",
        "cinder",
        "neutron",
        "nova",
        "barbican",
        "designate",
        "octavia",
    ],
    "openstack-extended": ["gnocchi", "ceilometer", "aodh", "senlin"],
}



import subprocess

from cliff.command import Command
from prompt_toolkit import prompt


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "--type",
            default="ssh",
            choices=["ansible", "clush", "container", "ssh"],
            help="Type of the console (default: %(default)s)",
        )
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Hostname or address of the console to connect",
        )
        return parser

    def take_action(self, parsed_args):
        type_console = parsed_args.type
        host = parsed_args.host[0]

        # If certain characters are contained in the hostname, then
        # enforce a certain console type.

        # ctl001/
        if host.endswith("/"):
            type_console = "container_prompt"
        # ctl001/rabbitmq
        elif "/" in host:
            type_console = "container"
        # .ctl001
        elif host.startswith("."):
            type_console = "ansible"
            host = host[1:]
        # :ctl00[1-3]
        elif host.startswith(":"):
            type_console = "clush"
            host = host[1:]

        ssh_options = "-o StrictHostKeyChecking=no -o LogLevel=ERROR"

        if type_console == "ansible":
            subprocess.call(f"/run-ansible-console.sh {host}", shell=True)
        elif type_console == "clush":
            subprocess.call(
                f"/usr/local/bin/clush -l dragon -w {host}",
                shell=True,
            )
        elif type_console == "ssh":
            # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
            subprocess.call(
                f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{host}",
                shell=True,
            )
        elif type_console == "container_prompt":
            while True:
                command = prompt(f"{host[:-1]}>>> ")
                if command in ["Exit", "exit", "EXIT"]:
                    break

                ssh_command = f"docker {command}"
                # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
                subprocess.call(
                    f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{host[:-1]} {ssh_command}",
                    shell=True,
                )
        elif type_console == "container":
            target_containername = host.split("/")[1]
            target_host = host.split("/")[0]
            target_command = "bash"

            ssh_command = f"docker exec -it {target_containername} {target_command}"
            ssh_options = (
                "-o RequestTTY=force -o StrictHostKeyChecking=no -o LogLevel=ERROR"
            )

            # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
            subprocess.call(
                f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{target_host} {ssh_command}",
                shell=True,
            )



from loguru import logger
from pottery import Redlock

from osism import utils


def for_device(name, parameters={}):

    device = utils.nb.dcim.devices.get(name=name)

    if (
        "device_type" not in device.custom_fields
        or device.custom_fields["device_type"] != "switch"
    ):
        return

    if "Managed by OSISM" not in [str(x) for x in device.tags]:
        return

    if "deployment_enabled" in device.custom_fields and not bool(
        device.custom_fields["deployment_enabled"]
    ):
        return

    if "deployment_type" not in device.custom_fields:
        return

    # Allow only one change per time
    lock = Redlock(
        key=f"lock_check_{name}", masters={utils.redis}, auto_release_time=120
    )
    lock.acquire()

    logger.info(
        f"Check configuration for device {device.name} with plugin {device.custom_fields['deployment_type']}"
    )

    deployment_type = device.custom_fields["deployment_type"]
    logger.error(
        f"Deployment type {deployment_type} for device {device.name} not supported"
    )
    last_configuration = None

    if last_configuration:
        for line in last_configuration.split("\n"):
            logger.info(f"configuration - {device.name}: {line}")

    lock.release()



import argparse
import subprocess

from cliff.command import Command
from prompt_toolkit import prompt


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument("host", nargs=1, type=str, help="Hostname or address")
        parser.add_argument(
            "command",
            nargs=argparse.REMAINDER,
            type=str,
            help="Command to run",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        command = " ".join(parsed_args.command)

        ssh_options = "-o StrictHostKeyChecking=no -o LogLevel=ERROR"

        if not command:
            while True:
                command = prompt(f"{host}>>> ")
                if command in ["Exit", "exit", "EXIT"]:
                    break

                ssh_command = f"docker {command}"
                # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
                subprocess.call(
                    f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{host} {ssh_command}",
                    shell=True,
                )
        else:
            ssh_command = f"docker {command}"

            # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
            subprocess.call(
                f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{host} {ssh_command}",
                shell=True,
            )



from pathlib import Path

import yaml

MAP_ROLE2ENVIRONMENT = {}

for path in Path("/interface/playbooks").glob("*-ansible.yml"):
    with open(path) as fp:
        data = yaml.load(fp, Loader=yaml.SafeLoader)

    MAP_ROLE2ENVIRONMENT = MAP_ROLE2ENVIRONMENT | data



import subprocess

from cliff.command import Command


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument("type", nargs=1, type=str, help="Type of the worker")
        return parser

    def take_action(self, parsed_args):
        queue = parsed_args.type[0]

        if queue in ["openstack", "netbox", "conductor"]:
            tasks = queue
        else:
            tasks = queue[:-8]

        # NOTE: use python interface in the future, works for the moment
        if tasks == "osism":
            p = subprocess.Popen(
                f"celery -A osism.tasks.ansible worker -n {queue} --loglevel=INFO -Q {queue}",
                shell=True,
            )
        else:
            p = subprocess.Popen(
                f"celery -A osism.tasks.{tasks} worker -n {queue} --loglevel=INFO -Q {queue}",
                shell=True,
            )
        p.wait()



import subprocess

from cliff.command import Command


class Images(Command):
    def get_parser(self, prog_name):
        parser = super(Images, self).get_parser(prog_name)

        # NOTE: This is a workaround. argparse.REMAINDER does not work well to pass all
        # arguments to openstack-image-manager. This is improved by switching from cliff
        # to typer. Then openstack-image-manager can simply be included directly at this
        # point.

        parser.add_argument(
            "--dry-run",
            default=False,
            help="Do not perform any changes",
            action="store_true",
        )
        parser.add_argument(
            "--hide",
            default=True,
            help="Hide images that should be deleted",
            action="store_true",
        )
        parser.add_argument(
            "--latest",
            default=False,
            help="Only import the latest version for images of type multi",
            action="store_true",
        )
        parser.add_argument(
            "--cloud", type=str, help="Cloud name in clouds.yaml", default="openstack"
        )
        parser.add_argument(
            "--hypervisor",
            type=str,
            help="Set hypervisor type meta information",
            default=None,
        )
        parser.add_argument(
            "--filter",
            type=str,
            help="Filter images with a regex on their name",
            default=None,
        )
        parser.add_argument(
            "--name",
            type=str,
            action="append",
            help="Name of the image to process, use repeatedly for multiple images",
        )

        return parser

    def take_action(self, parsed_args):
        cloud = parsed_args.cloud
        dry_run = parsed_args.dry_run
        filter = parsed_args.filter
        hide = parsed_args.hide
        hypervisor = parsed_args.hypervisor
        latest = parsed_args.latest
        names = parsed_args.name

        arguments = []
        if cloud:
            arguments.append(f"--cloud '{cloud}'")
        if filter:
            arguments.append(f"--filter '{filter}'")
        if dry_run:
            arguments.append("--dry-run")
        if latest:
            arguments.append("--latest")
        if hide:
            arguments.append("--hide")
        if hypervisor:
            arguments.append(f"--hypervisor '{hypervisor}'")
        if names:
            for name in names:
                arguments.append(f"--name '{name}'")

        joined_arguments = " ".join(arguments)
        subprocess.call(
            f"/usr/local/bin/openstack-image-manager --images=/etc/images {joined_arguments}",
            shell=True,
        )



__all__ = ["__version__"]

import pbr.version

version_info = pbr.version.VersionInfo("osism")
# We have a circular import problem when we first run python setup.py sdist
# It's harmless, so deflect it.
try:
    __version__ = version_info.version_string()
except AttributeError:
    __version__ = None



import os


# Read secret from file
def read_secret(secret_name):
    try:
        f = open("/run/secrets/" + secret_name, "r", encoding="utf-8")
    except EnvironmentError:
        return ""
    else:
        with f:
            return f.readline().strip()


NETBOX_URL = os.getenv("NETBOX_API")
NETBOX_TOKEN = os.getenv("NETBOX_TOKEN", read_secret("NETBOX_TOKEN"))
IGNORE_SSL_ERRORS = os.getenv("IGNORE_SSL_ERRORS", "True") == "True"

BASE_PATH = os.getenv("BASE_PATH", "/devicetype-library/device-types/")
VENDORS = os.getenv("VENDORS", "").split()



# Copyright OSISM GmbH, 2022-2023
# LICENSE: CC BY-NC 4.0

import glob
import os

from loguru import logger
from pottery import Redlock
import pynetbox
import yaml

from osism import utils
from osism.actions import generate_configuration, deploy_configuration


def load_data_from_filesystem(collection=None, device=None, state=None):
    """Loads all known data for a given device or an entire collection
    from the file system (/netbox) in a given state.
    """

    if not state or state in ["0", "None"]:
        state = "a"

    data = {}
    if not device:
        logger.info(f"Loading collection {collection}")

        if os.path.isfile("/netbox/{CONF.collection}/{CONF.state}.yaml"):
            with open(f"/netbox/{collection}/{state}.yaml") as fp:
                data = yaml.load(fp, Loader=yaml.SafeLoader)

        for directory in glob.glob(f"/netbox/{collection}/*/"):
            with open(f"{directory}{state}.yaml") as fp:
                data_a = yaml.load(fp, Loader=yaml.SafeLoader)
            # data = data | data_a
            data = {**data_a, **data}

    elif device and collection:
        if not os.path.isfile(
            "/netbox/{CONF.collection}/{CONF.device}/{CONF.state}.yaml"
        ):
            logger.error(
                f"State {state} for device {device} in collection {collection} is not available"
            )
            return data

        logger.info(f"Loading device {device} from collection {collection}")

        with open(f"/netbox/{collection}/{device}/{state}.yaml") as fp:
            data = yaml.load(fp, Loader=yaml.SafeLoader)

    elif device:
        # Try to find the collection of the specified device
        # A device can be in exactly one collection
        result = [x[0] for x in os.walk("/netbox") if device in x[0]]
        if result:
            logger.info(f"Loading device {device}")

            try:
                with open(f"{result[0]}/{state}.yaml") as fp:
                    data = yaml.load(fp, Loader=yaml.SafeLoader)
            except:  # noqa
                logger.error(f"State {state} for device {device} is not available")
                return data
        else:
            logger.error(f"Device {device} is not defined in any collection")

    else:
        logger.error("Specify at least a collection or a device")

    return data


def get_state(device):
    """Gets the state (device_state) stored in the Netbox for a  device."""

    result = None
    device_a = utils.nb.dcim.devices.get(name=device)
    result = device_a.custom_fields["device_state"]

    return result


def get_states(devices):
    """Gets the state (device_state) stored in the Netbox for a list of devices."""

    result = {}
    for device in devices:
        device_a = utils.nb.dcim.devices.get(name=device)
        result[device] = device_a.custom_fields["device_state"]

    return result


def get_transitions(devices):
    """Gets the transition (device_transition) stored in the Netbox for a list of devices."""

    result = {}
    for device in devices:
        device_a = utils.nb.dcim.devices.get(name=device)
        result[device] = device_a.custom_fields["device_transition"]

    return result


def get_lag_interfaces(data):
    """Returns all defined LAGs."""

    result = {}
    for device in data:
        result[device] = []
        for interface in data[device]:
            if data[device][interface]["type"] == "port-channel":
                for interface in data[device][interface]["interfaces"]:
                    if interface not in result[device]:
                        result[device].append(interface)

    return result


def manage_interfaces(device, data):
    """Manage interfaces."""
    primary_address = None
    lag_interfaces = get_lag_interfaces(data)

    for interface in data[device]:

        if data[device][interface]["type"] in ["virtual", "port-channel", "mlag"]:
            continue

        device_a = utils.nb.dcim.devices.get(name=device)
        device_target = data[device][interface]["device"]
        device_b = utils.nb.dcim.devices.get(name=device_target)

        interface_a = utils.nb.dcim.interfaces.get(name=interface, device=device)
        interface_b = utils.nb.dcim.interfaces.get(
            name=data[device][interface]["interface"],
            device=data[device][interface]["device"],
        )

        if not interface_a:
            logger.error(f"{device} # {interface} --> not found")

        if not interface_b:
            logger.error(
                f"{data[device][interface]['device']} # {data[device][interface]['interface']} --> not found"
            )

        # Ignore interfaces without an mac address
        try:
            if "mac_address" in data[device][interface]:
                interface_a.mac_address = data[device][interface]["mac_address"]
                interface_a.save()
            elif interface_a.mac_address:
                interface_a.mac_address = None
                interface_a.save()
        except:  # noqa E722
            pass

        if interface_a and "data" in data[device][interface]:
            interface_a.update(data[device][interface]["data"])

            if "enabled" in data[device][interface]["data"] and interface_b:
                interface_b.enabled = bool(data[device][interface]["data"]["enabled"])

        # Add all addresses to the interface
        if "addresses" in data[device][interface]:
            for address in data[device][interface]["addresses"]:
                address_a = utils.nb.ipam.ip_addresses.get(address=address)
                if type(address) == str:
                    address_a = utils.nb.ipam.ip_addresses.get(address=address)
                    logger.info(f"Address {address} -> {interface}")
                    if not address_a:
                        utils.nb.ipam.ip_addresses.create(
                            address=address,
                            assigned_object_type="dcim.interface",
                            assigned_object_id=interface_a.id,
                        )
                else:
                    address_a = utils.nb.ipam.ip_addresses.get(
                        address=address["address"]
                    )
                    logger.info(f"Address {address['address']} -> {interface}")
                    if not address_a:
                        address_a = utils.nb.ipam.ip_addresses.create(
                            assigned_object_type="dcim.interface",
                            assigned_object_id=interface_a.id,
                            **address,
                        )
                    if "primary" in address and bool(address["primary"]):
                        primary_address = address_a.id
                        device_a.primary_ip4 = address_a.id
                        device_a.save()

        # Remove addresses from the interface that have been removed
        for address in utils.nb.ipam.ip_addresses.filter(
            device=device, interface=interface
        ):
            delete = True
            if "addresses" in data[device][interface]:
                for address_a in data[device][interface]["addresses"]:
                    if type(address_a) == str and address_a == str(address):
                        delete = False
                    elif "address" in address_a and address_a["address"] == str(
                        address
                    ):
                        delete = False

            if delete:
                address.delete()

        logger.info(f"{interface_a} -> {device_target} # {interface_b}")

        if interface_a.label:
            port_a = interface_a.label
        # EthernetXX/Y
        elif "Ethernet" in interface_a.name and "/" in interface_a.name:
            port_a = interface_a.name[8:].split("/")[0]
        # EthernetXX
        elif "Ethernet" in interface_a.name:
            port_a = interface_a.name[8:]
        # etherXX
        elif "ether" in interface_a.name:
            port_a = interface_a.name[5:]
        # ethXX
        elif "eth" in interface_a.name:
            port_a = interface_a.name[3:]
        # sfp-sfpplusXX
        elif "sfp-sfpplus" in interface_a.name:
            port_a = interface_a.name[11:]
        # qsfp-qsfpplusXX
        elif "qsfp-qsfpplus" in interface_a.name:
            port_a = interface_a.name[13:]
        # qsfpplusXX
        elif "qsfpplus" in interface_a.name:
            port_a = interface_a.name[8:]
        else:
            port_a = interface_a.name

        if interface_b.label:
            port_b = interface_b.label
        # EthernetXX/Y
        elif "Ethernet" in interface_b.name and "/" in interface_b.name:
            port_b = interface_b.name[8:].split("/")[0]
        # EthernetXX
        elif "Ethernet" in interface_b.name:
            port_b = interface_b.name[8:]
        # etherXX
        elif "ether" in interface_b.name:
            port_b = interface_b.name[5:]
        # ethXX
        elif "eth" in interface_b.name:
            port_b = interface_b.name[3:]
        # sfp-sfpplusXX
        elif "sfp-sfpplus" in interface_b.name:
            port_b = interface_b.name[11:]
        # qsfp-qsfpplusXX
        elif "qsfp-qsfpplus" in interface_b.name:
            port_b = interface_b.name[13:]
        # qsfpplusXX
        elif "qsfpplus" in interface_b.name:
            port_b = interface_b.name[8:]
        else:
            port_b = interface_b.name

        try:
            position_a = int(device_a.position)
        except:
            # NOTE: dirty workaround so that it works for the moment also for nodes without
            #       a position in housings
            position_a = 999

        try:
            position_b = int(device_b.position)
        except:
            # NOTE: dirty workaround so that it works for the moment also for nodes without
            #       a position in housings
            position_b = 999

        near_end_a = f"{position_a}:{port_a}"
        if device_a.rack.name == device_b.rack.name:
            far_end_a = f"{position_b}:{port_b}"
        else:
            far_end_a = f"{device_b.rack.name}-{position_b}:{port_b}"
        label_a = f"{near_end_a} / {far_end_a}"

        near_end_b = f"{position_b}:{port_b}"
        if device_b.rack.name == device_a.rack.name:
            far_end_b = f"{position_a}:{port_a}"
        else:
            far_end_b = f"{device_a.rack.name}-{position_a}:{port_a}"
        label_b = f"{near_end_b} / {far_end_b}"

        interface_a.update({"description": label_a})
        interface_b.update({"description": label_b})

        termination_a = {"object_type": "dcim.interface", "object_id": interface_a.id}
        termination_b = {"object_type": "dcim.interface", "object_id": interface_b.id}

        try:
            connection = utils.nb.dcim.cables.create(
                a_terminations=[termination_a],
                b_terminations=[termination_b],
                type=data[device][interface]["type"],
            )
        except pynetbox.core.query.RequestError as e:
            # The "Duplicate termination found" error can be ignored
            if "Duplicate termination found" not in e.error:
                logger.error(f"ERROR --> {e.error}")
            pass

        # ensure that all interfaces are enabled that should be enabled
        if not interface_a.enabled:

            if "enabled" in data[device][interface]["data"]:
                interface_a.enabled = bool(data[device][interface]["data"]["enabled"])
            else:
                interface_a.enabled = True

            if interface_a.enabled:
                logger.info(f"{device_a} # {interface_a} --> enabled")
            else:
                logger.info(f"{device_a} # {interface_a} --> disabled")

            interface_a.save()

        if not interface_b.enabled:
            if (
                "data" in data[device][interface]
                and "enabled" in data[device][interface]["data"]
            ):
                interface_b.enabled = bool(data[device][interface]["data"]["enabled"])
            else:
                interface_b.enabled = True

            if interface_b.enabled:
                logger.info(f"{device_b} # {interface_b} --> enabled")
            else:
                logger.info(f"{device_b} # {interface_b} --> disabled")

            interface_b.save()

        if "vlans" in data[device][interface]:
            tagged = False
            interface_a.untagged_vlan = None
            interface_a.tagged_vlans = []
            for vlan in data[device][interface]["vlans"]:
                vlan_a = utils.nb.ipam.vlans.get(vid=vlan)
                if not vlan_a:
                    try:
                        vlan_a = utils.nb.ipam.vlans.create(
                            name=f"VLAN {vlan}", vid=vlan
                        )
                    except pynetbox.core.query.RequestError as e:
                        logger.error(f"ERROR --> {e}")
                        pass

                if data[device][interface]["vlans"][vlan] == "untagged":
                    logger.info(f"Untagged VLAN {vlan_a.vid} -> {interface_a.name}")
                    interface_a.untagged_vlan = vlan_a.id

                    if interface_a.name not in lag_interfaces[device]:
                        logger.info(f"Tagged VLAN {vlan_a.vid} -> {interface_b.name}")
                        interface_b.untagged_vlan = vlan_a.id

                elif vlan_a.id not in interface_a.tagged_vlans:
                    logger.info(f"Tagged VLAN {vlan_a.vid} -> {interface_a.name}")
                    interface_a.tagged_vlans.append(vlan_a.id)

                    if interface_a.name not in lag_interfaces[device]:
                        logger.info(f"Tagged VLAN {vlan_a.vid} -> {interface_b.name}")
                        interface_b.tagged_vlans.append(vlan_a.id)

                    tagged = True

            if tagged:
                interface_a.mode = "tagged"

                if interface_a.name not in lag_interfaces[device]:
                    interface_b.mode = "tagged"
            else:
                interface_a.mode = "access"

                if interface_a.name not in lag_interfaces[device]:
                    interface_b.mode = "access"

            interface_a.save()

            if interface_a.name not in lag_interfaces[device]:
                interface_b.save()

    # Remove the primary IP address if it is no longer set
    if not primary_address:
        device_a.primary_ip4 = None
        device_a.save()


def manage_port_channels(device, data):
    """Manage port channels (not MLAGs)."""

    for interface in data[device]:
        if data[device][interface]["type"] == "port-channel":
            logger.info(
                f"Local port channel {device} # {interface} -> {data[device][interface]['interfaces']}"
            )
            device_a = utils.nb.dcim.devices.get(name=device)

            # Create the local port channel
            port_channel_a = utils.nb.dcim.interfaces.get(name=interface, device=device)
            if not port_channel_a:
                try:
                    port_channel_a = utils.nb.dcim.interfaces.create(
                        name=interface, device=device_a.id, type="lag"
                    )
                except pynetbox.core.query.RequestError as e:
                    logger.error(f"ERROR --> {e}")
                    pass

            # Create the remote port channels and add the local interfaces to the local port channel
            remote_port_channels = []
            for interface_x in data[device][interface]["interfaces"]:
                interface_a = utils.nb.dcim.interfaces.get(
                    name=interface_x, device=device
                )

                # NOTE: The VLANs on the Ethernet interfaces on the local devices are preserved for
                #       visibility in the Netbox.
                # interface_a.untagged_vlan = None
                # interface_a.tagged_vlans = []

                interface_a.lag = port_channel_a
                interface_a.save()

                port_channel_b_name = (
                    f"Port-Channel{data[device][interface]['channel']}"
                )
                interface_b = utils.nb.dcim.interfaces.get(
                    name=interface_a.connected_endpoint.name,
                    device=interface_a.connected_endpoint.device,
                )

                interface_b.untagged_vlan = None
                interface_b.tagged_vlans = []

                logger.info(
                    f"Remote port channel {interface_b.device.name} # {port_channel_b_name} -> {interface_b.device.name} # {interface_b.name} ({interface_a.name})"
                )

                port_channel_b = utils.nb.dcim.interfaces.get(
                    name=port_channel_b_name, device=interface_b.device
                )
                if not port_channel_b:
                    try:
                        port_channel_b = utils.nb.dcim.interfaces.create(
                            name=port_channel_b_name,
                            device=interface_b.device.id,
                            type="lag",
                        )
                    except pynetbox.core.query.RequestError as e:
                        logger.error(f"ERROR --> {e}")
                        pass

                interface_b.lag = port_channel_b
                interface_b.save()

                remote_port_channels.append(port_channel_b)

            # Assign IP addresses to the local port channel
            if "addresses" in data[device][interface]:
                for address in data[device][interface]["addresses"]:
                    address_a = utils.nb.ipam.ip_addresses.get(address=address)
                    if type(address) == str:
                        address_a = utils.nb.ipam.ip_addresses.get(address=address)
                        logger.info(f"Address {address} -> {interface}")
                        if not address_a:
                            utils.nb.ipam.ip_addresses.create(
                                address=address,
                                assigned_object_type="dcim.interface",
                                assigned_object_id=port_channel_a.id,
                            )
                    else:
                        address_a = utils.nb.ipam.ip_addresses.get(
                            address=address["address"]
                        )
                        logger.info(f"Address {address['address']} -> {interface}")
                        if not address_a:
                            address_a = utils.nb.ipam.ip_addresses.create(
                                assigned_object_type="dcim.interface",
                                assigned_object_id=port_channel_a.id,
                                **address,
                            )
                        if "primary" in address and bool(address["primary"]):
                            device_a.primary_ip4 = address_a.id
                            device_a.save()

            # Remove addresses from the local port channel that have been removed
            for address in utils.nb.ipam.ip_addresses.filter(
                device=device, interface=interface
            ):
                delete = True
                if "addresses" in data[device][interface]:
                    for address_a in data[device][interface]["addresses"]:
                        if type(address_a) == str and address_a == str(address):
                            delete = False
                        elif "address" in address_a and address_a["address"] == str(
                            address
                        ):
                            delete = False

                if delete:
                    address.delete()

            # Assign VLANs to the local port channel as well as the remote port channels
            port_channel_a.untagged_vlan = None
            port_channel_a.tagged_vlans = []
            port_channel_b.untagged_vlan = None
            port_channel_b.tagged_vlans = []

            if "vlans" in data[device][interface]:
                tagged = False
                for vlan in data[device][interface]["vlans"]:
                    vlan_a = utils.nb.ipam.vlans.get(vid=vlan)
                    if not vlan_a:
                        try:
                            vlan_a = utils.nb.ipam.vlans.create(
                                name=f"VLAN {vlan}", vid=vlan
                            )
                        except pynetbox.core.query.RequestError as e:
                            logger.error(f"ERROR --> {e}")
                            pass

                    if data[device][interface]["vlans"][vlan] == "untagged":
                        logger.info(
                            f"Untagged VLAN {vlan_a.vid} -> {port_channel_a.name}"
                        )
                        port_channel_a.untagged_vlan = vlan_a.id

                        for port_channel_b in remote_port_channels:
                            logger.info(
                                f"Untagged VLAN {vlan_a.vid} -> {port_channel_b.name}"
                            )
                            port_channel_b.untagged_vlan = vlan_a.id
                    elif vlan_a.id not in port_channel_a.tagged_vlans:
                        logger.info(
                            f"Tagged VLAN {vlan_a.vid} -> {port_channel_a.name}"
                        )
                        port_channel_a.tagged_vlans.append(vlan_a.id)
                        tagged = True

                        for port_channel_b in remote_port_channels:
                            logger.info(
                                f"Tagged VLAN {vlan_a.vid} -> {port_channel_b.name}"
                            )
                            port_channel_b.tagged_vlans.append(vlan_a.id)

                if tagged:
                    port_channel_a.mode = "tagged"

                    for port_channel_b in remote_port_channels:
                        port_channel_b.mode = "tagged"
                else:
                    port_channel_a.mode = "access"

                    for port_channel_b in remote_port_channels:
                        port_channel_b.mode = "access"

            port_channel_a.save()
            port_channel_b.save()


def remove_port_channels(device, data):
    """Remove local and remote port channels that no longer exist."""

    for interface in utils.nb.dcim.interfaces.filter(device=device, type="lag"):
        delete = True
        for interface_a in data[device]:
            if (
                data[device][interface_a]["type"] == "port-channel"
                and str(interface) == interface_a
            ):
                delete = False

        if delete and "Port-Channel" not in interface.name:
            members = utils.nb.dcim.interfaces.filter(lag_id=interface.id)
            for member in members:
                member.connected_endpoint.lag.delete()
            interface.delete()


def manage_virtual_interfaces(device, data):
    """Manage virtual interfaces."""

    for interface in data[device]:
        if data[device][interface]["type"] == "virtual":
            logger.info(f"Virtual interface {interface} for {device}")

            device_a = utils.nb.dcim.devices.get(name=device)

            interface_a = utils.nb.dcim.interfaces.get(name=interface, device=device)
            if not interface_a:
                try:
                    interface_a = utils.nb.dcim.interfaces.create(
                        name=interface,
                        device=device_a.id,
                        type="virtual",
                        **data[device][interface]["data"],
                    )
                except pynetbox.core.query.RequestError as e:
                    logger.error(f"ERROR --> {e}")
                    pass

            if "addresses" in data[device][interface]:
                for address in data[device][interface]["addresses"]:
                    address_a = utils.nb.ipam.ip_addresses.get(address=address)
                    if type(address) == str:
                        address_a = utils.nb.ipam.ip_addresses.get(address=address)
                        logger.info(f"Address {address} -> {interface}")
                        if not address_a:
                            utils.nb.ipam.ip_addresses.create(
                                address=address,
                                assigned_object_type="dcim.interface",
                                assigned_object_id=interface_a.id,
                            )
                    else:
                        address_a = utils.nb.ipam.ip_addresses.get(
                            address=address["address"]
                        )
                        logger.info(f"Address {address['address']} -> {interface}")
                        if not address_a:
                            address_a = utils.nb.ipam.ip_addresses.create(
                                assigned_object_type="dcim.interface",
                                assigned_object_id=interface_a.id,
                                **address,
                            )
                        if "primary" in address and bool(address["primary"]):
                            device_a.primary_ip4 = address_a.id
                            device_a.save()

            # Remove addresses from the interface that have been removed
            for address in utils.nb.ipam.ip_addresses.filter(
                device=device, interface=interface
            ):
                delete = True
                if "addresses" in data[device][interface]:
                    for address_a in data[device][interface]["addresses"]:
                        if type(address_a) == str and address_a == str(address):
                            delete = False
                        elif "address" in address_a and address_a["address"] == str(
                            address
                        ):
                            delete = False

                if delete:
                    address.delete()

            if "vlans" in data[device][interface]:
                tagged = False
                interface_a.untagged_vlan = None
                interface_a.tagged_vlans = []
                for vlan in data[device][interface]["vlans"]:
                    vlan_a = utils.nb.ipam.vlans.get(vid=vlan)
                    if not vlan_a:
                        try:
                            vlan_a = utils.nb.ipam.vlans.create(
                                name=f"VLAN {vlan}", vid=vlan
                            )
                        except pynetbox.core.query.RequestError as e:
                            logger.error(f"ERROR --> {e}")
                            pass

                    if data[device][interface]["vlans"][vlan] == "untagged":
                        interface_a.untagged_vlan = vlan_a.id
                    elif vlan_a.id not in interface_a.tagged_vlans:
                        interface_a.tagged_vlans.append(vlan_a.id)
                        tagged = True

                if tagged:
                    interface_a.mode = "tagged"
                else:
                    interface_a.mode = "access"
                interface_a.save()


def remove_virtual_interfaces(device, data):
    """Remove virtual interfaces that no longer exist."""

    for interface in utils.nb.dcim.interfaces.filter(device=device, type="virtual"):
        delete = True
        for interface_a in data[device]:
            if (
                data[device][interface_a]["type"] == "virtual"
                and str(interface) == interface_a
            ):
                delete = False

        if delete:
            interface.delete()


def manage_mlag_devices(device, data):
    """Manage MLAG devices (not port channels)."""

    for interface in data[device]:
        if data[device][interface]["type"] == "mlag":
            data_a = data[device][interface]["data"]
            device_a = utils.nb.dcim.devices.get(name=device)

            logger.info(
                f"Local port channel {device} # Port-Channel{data_a['channel']}"
            )

            port_channel_a = utils.nb.dcim.interfaces.get(
                name=f"Port-Channel{data_a['channel']}", device=device
            )
            if not port_channel_a:
                try:
                    port_channel_a = utils.nb.dcim.interfaces.create(
                        name=f"Port-Channel{data_a['channel']}",
                        device=device_a.id,
                        type="lag",
                    )
                except pynetbox.core.query.RequestError as e:
                    logger.error(f"ERROR --> {e}")
                    pass

            for interface_x in data[device][interface]["interfaces"]:
                interface_a = utils.nb.dcim.interfaces.get(
                    name=interface_x, device=device
                )
                interface_a.lag = port_channel_a
                interface_a.save()

            logger.info(f"Virtual interface {data_a['vlan']} for {device}")
            interface_a = utils.nb.dcim.interfaces.get(
                name=f"Vlan{data_a['vlan']}", device=device
            )
            if not interface_a:
                try:
                    interface_a = utils.nb.dcim.interfaces.create(
                        name=f"Vlan{data_a['vlan']}", device=device_a.id, type="virtual"
                    )
                except pynetbox.core.query.RequestError as e:
                    logger.error(f"ERROR --> {e}")
                    pass

            vlan_a = utils.nb.ipam.vlans.get(vid=data_a["vlan"])
            interface_a.untagged_vlan = vlan_a
            interface_a.parent = port_channel_a
            interface_a.save()

            # logger.info(f"Address {data_a['address']} -> {interface_a.name}")
            address_a = utils.nb.ipam.ip_addresses.get(address=data_a["address"])
            if not address_a:
                utils.nb.ipam.ip_addresses.create(
                    address=data_a["address"],
                    assigned_object_type="dcim.interface",
                    assigned_object_id=interface_a.id,
                )


def set_maintenance(device, state):
    """Set the maintenance state for a device in the Netbox."""

    logger.info(f"Set maintenance state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"maintenance": state}
    device_a.save()


def set_state(device, state, state_type):
    """Set the state for a device in the Netbox."""

    lock = Redlock(key=f"lock_state_{device}", masters={utils.redis})
    lock.acquire()

    if state_type == "power":
        set_power_state(device, state)
    elif state_type == "provision":
        set_provision_state(device, state)
    elif state_type == "introspection":
        set_introspection_state(device, state)
    elif state_type == "ironic":
        set_ironic_state(device, state)
    elif state_type == "deployment":
        set_deployment_state(device, state)
    else:
        set_device_state(device, state)

    lock.release()


def set_provision_state(device, state):
    """Set the provision state (provision_state) for a device in the Netbox."""

    logger.info(f"Set provision state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"provision_state": state}
    device_a.save()


def set_ironic_state(device, state):
    """Set the ironic state (ironic_state) for a device in the Netbox."""

    logger.info(f"Set ironic state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"ironic_state": state}
    device_a.save()


def set_introspection_state(device, state):
    """Set the introspection state (introspection_state) for a device in the Netbox."""

    logger.info(f"Set introspection state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"introspection_state": state}
    device_a.save()


def set_deployment_state(device, state):
    """Set the deployment state (deployment_state) for a device in the Netbox."""

    logger.info(f"Set deployment state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"deployment_state": state}
    device_a.save()


def set_power_state(device, state):
    """Set the power state (power_state) for a device in the Netbox."""

    logger.info(f"Set power state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"power_state": state}
    device_a.save()


def set_device_state(device, state):
    """Set the state (device_state) for a device in the Netbox."""

    logger.info(f"Set state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"device_state": state}
    device_a.save()


def set_device_transition(device, transition):
    """Set the transition (device_transition) for a device in the Netbox."""

    logger.info(f"Set transition of device {device} = {transition}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"device_transition": transition}
    device_a.save()


def get_device_state(device, states):
    """Get the state (device_state) for a device in the Netbox."""

    return states[device]


def get_device_transition(device, transitions):
    """Get the transition (device_transition) for a device in the Netbox."""

    return transitions[device]


def get_connected_devices(device, data):
    """Get all devices that are connected to a device in a certain state."""

    result = []

    if device not in data:
        return result

    for interface in data[device]:
        if "device" in data[device][interface]:
            result.append(data[device][interface]["device"])

    return result


def run(device, state=None, data={}, enforce=False):
    """Transition a device to a specific state."""

    # If no state is specified use the state that is stored in the Netbox
    if not state or state in ["0", "None"]:
        state = get_state(device)

        # If the state in the Netbox is 0/None then set the state to a
        if state in ["0", "None"]:
            state = "a"

    if not data:
        data = load_data_from_filesystem(None, device, state)

    states = get_states(data.keys())
    current_state = get_device_state(device, states)

    # Device is already in the target state, no transition necessary
    if not enforce and current_state == state:
        logger.info(f"Device {device} is already in state {state}")
        return

    transitions = get_transitions(data.keys())
    current_transition = get_device_transition(device, transitions)

    if current_state and current_state not in ["0", "None"]:
        current_data = load_data_from_filesystem(None, device, current_state)
    else:
        current_data = {}

    # One transition is already running, no second transition possible
    if not enforce and current_transition and current_transition != "0":
        logger.info(f"{device} is already in transit")
        return

    # Get connected devices in source and target state
    connected_devices = set(
        get_connected_devices(device, current_data)
        + get_connected_devices(device, data)
    )
    logger.info(connected_devices)

    # Allow only one active transition per device
    lock = Redlock(key=f"lock_{device}", masters={utils.redis}, auto_release_time=120)
    lock.acquire()

    # transition: from-to, phase 1 (modifications in the Netbox)
    transition = f"from_{states[device]}-to_{state}-phase_1"
    set_device_transition(device, transition)

    manage_interfaces(device, data)
    manage_port_channels(device, data)
    remove_port_channels(device, data)
    manage_virtual_interfaces(device, data)
    remove_virtual_interfaces(device, data)
    manage_mlag_devices(device, data)

    set_device_state(device, f"{state}-phase_1")

    # transition: from-to, phase 2 (generate the new configuration)
    transition = f"from_{states[device]}-to_{state}-phase_2"
    set_device_transition(device, transition)

    for connected_device in [x for x in connected_devices if x]:
        generate_configuration.for_device(connected_device)

    set_device_state(device, f"{state}-phase_2")

    # transition: from-to, phase 3 (deploy the new configuration)
    transition = f"from_{states[device]}-to_{state}-phase_3"

    for connected_device in [x for x in connected_devices if x]:
        deploy_configuration.for_device(connected_device)

    set_device_transition(device, transition)
    set_device_state(device, f"{state}-phase_3")

    # transition: from-to, phase 4 (validate the deployed configuration)
    transition = f"from_{states[device]}-to_{state}-phase_4"
    set_device_transition(device, transition)

    # for connected_device in connected_devices:
    #     validate_configuration.for_device(connected_device)

    set_device_state(device, f"{state}-phase_4")

    # target state reached
    transition = ""
    set_device_transition(device, transition)
    set_device_state(device, f"{state}")

    lock.release()



import argparse
import time

from celery import group
from celery.result import GroupResult
from cliff.command import Command
from loguru import logger
from tabulate import tabulate

from osism.core import enums
from osism.core.playbooks import MAP_ROLE2ENVIRONMENT
from osism.tasks import ansible, ceph, kolla
from osism.utils import redis


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "--environment",
            "-e",
            type=str,
            help="Environment that is to be used explicitly",
        )
        parser.add_argument(
            "--overwrite",
            type=str,
            help="Overwrite the environment after the mapping on the worker (has no effect on ceph and kolla environments)",
        )
        parser.add_argument(
            "--sub",
            type=str,
            help="Use a sub-environment (e.g. ceph.zone-b or kolla.zone-b)",
        )
        parser.add_argument(
            "--action",
            "-a",
            dest="action",
            type=str,
            help="Action to be applied (can only be used for OpenStack playbooks) (default: %(default)s)",
            choices=[
                "deploy",
                "precheck",
                "pull",
                "reconfigure",
                "refresh-containers",
                "rolling-upgrade",
                "stop",
                "upgrade",
            ],
            default="deploy",
        )
        parser.add_argument("role", nargs="?", type=str, help="Role to be applied")
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--format",
            default="log",
            help="Output type",
            const="log",
            nargs="?",
            choices=["script", "log"],
        ),
        parser.add_argument(
            "--timeout",
            default=300,
            type=int,
            help="Timeout to end if there is no output",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the role has been applied",
            action="store_true",
        )
        return parser

    def _handle_loadbalancer(self, t, wait, format, timeout):
        # process the parent task
        rc = self._handle_task(t.parent, wait, format, timeout)

        # It is necessary to wait for all task even if this is not excpected by the
        # user because of the following exception thrown by the garbage collector.
        #
        # Exception ignored in: <function AsyncResult.__del__ at 0x7f8c91ac74c0>
        # Traceback (most recent call last):
        # [...]
        # ImportError: sys.meta_path is None, Python is likely shutting down

        if not wait:
            t.parent.get()

        # process the child tasks
        if format == "log":
            for c in t.children:
                logger.info(
                    f"Task {c.task_id} is running in background. No more output. Check ARA for logs."
                )

        # As explained above, it is neceesary to wait for all tasks.
        t.get()

        return rc

    def _handle_task(self, t, wait, format, timeout):
        rc = 0
        if wait:
            p = redis.pubsub()
            p.subscribe(f"{t.task_id}")

            stoptime = time.time() + timeout
            while time.time() < stoptime:
                m = p.get_message(timeout=stoptime - time.time())
                if m:
                    stoptime = time.time() + timeout
                    if type(m["data"]) == bytes:
                        line = m["data"].decode("utf-8")
                        if line.startswith("RC: "):
                            rc = int(line[4:])
                            continue
                        if line == "QUIT":
                            redis.close()
                            # NOTE: Use better solution
                            return rc
                        print(line, end="")
                else:
                    logger.info(
                        f"No further output after {timeout} seconds. Therefore finish."
                    )
                    return rc

        else:
            if format == "log":
                logger.info(
                    f"Task {t.task_id} is running in background. No more output. Check ARA for logs."
                )
            elif format == "script":
                print(f"{t.task_id}")

            return rc

    def handle_role(
        self,
        arguments,
        environment,
        overwrite,
        sub,
        role,
        action,
        wait,
        format,
        timeout,
    ):
        # There is a special playbook ceph-ceph which should be called
        # with ceph. Therefore, the environment is set explicitly in
        # this case.
        if role == "ceph":
            environment = "ceph"

        if not environment:
            try:
                environment = MAP_ROLE2ENVIRONMENT[role]
            except:  # noqa: E722
                environment = "custom"

        if environment == "ceph":
            if sub:
                environment = f"{environment}.{sub}"
            if role.startswith("ceph-"):
                t = ceph.run.delay(environment, role[5:], arguments)
            else:
                t = ceph.run.delay(environment, role, arguments)
        elif environment == "kolla":
            if sub:
                environment = f"{environment}.{sub}"
            if action == "rolling-upgrade":
                action = "rolling_upgrade"
            kolla_arguments = [f"-e kolla_action={action}"] + arguments
            if role.startswith("kolla-"):
                t = kolla.run.delay(environment, role[6:], kolla_arguments)
            else:
                t = kolla.run.delay(environment, role, kolla_arguments)
        elif role == "loadbalancer-ng":
            if sub:
                environment = f"{environment}.{sub}"
            g = group(
                kolla.run.si(environment, playbook, arguments)
                for playbook in enums.LOADBALANCER_PLAYBOOKS
            )
            t = (
                kolla.run.s(environment, "loadbalancer-ng", arguments) | g
            ).apply_async()
        else:
            # Overwrite the environment
            if overwrite:
                environment = overwrite
            t = ansible.run.delay(environment, role, arguments)

        if isinstance(t, GroupResult):
            rc = self._handle_loadbalancer(t, wait, format, timeout)
        else:
            rc = self._handle_task(t, wait, format, timeout)

        return rc

    def take_action(self, parsed_args):
        action = parsed_args.action
        arguments = parsed_args.arguments
        environment = parsed_args.environment
        format = parsed_args.format
        overwrite = parsed_args.overwrite
        role = parsed_args.role
        sub = parsed_args.sub
        timeout = parsed_args.timeout
        wait = not parsed_args.no_wait

        rc = 0

        if not role:
            table = []
            for role in MAP_ROLE2ENVIRONMENT:
                table.append([role, MAP_ROLE2ENVIRONMENT[role]])
            logger.info(
                "No role given for execution. The roles listed in the table can be used."
            )
            print(tabulate(table, headers=["Role", "Environment"], tablefmt="psql"))
        elif role in enums.MAP_ROLE2ROLE:
            for r in enums.MAP_ROLE2ROLE[role]:
                rc = self.handle_role(
                    arguments,
                    environment,
                    overwrite,
                    sub,
                    action,
                    r,
                    wait,
                    format,
                    timeout,
                )
                if rc != 0:
                    break
        else:
            rc = self.handle_role(
                arguments,
                environment,
                overwrite,
                sub,
                role,
                action,
                wait,
                format,
                timeout,
            )

        return rc
